@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------------
    libconfig - A library for processing structured configuration files
-   Copyright (C) 2005-2015  Mark A Lindner
+   Copyright (C) 2005-2018  Mark A Lindner
 
    This file is part of libconfig.
 
@@ -44,23 +44,18 @@
 #include "wincompat.h"
 #include "grammar.h"
 #include "scanner.h"
+#include "util.h"
 
 #define PATH_TOKENS ":./"
 #define CHUNK_SIZE 16
-#define FLOAT_PRECISION 2
-
-#define TAB2DIGITS(x)  (((x) & 0xFF00) >> 8)
-#define DIGITS2TAB(x)  (((x) & 0xFF) << 8)
-#define TAB2TAB(x)     ((x) & 0xF)
-
-#define _new(T) (T *)calloc(1, sizeof(T)) /* zeroed */
-#define _delete(P) free((void *)(P))
+#define DEFAULT_TAB_WIDTH 2
+#define DEFAULT_FLOAT_PRECISION 6
 
 /* ------------------------------------------------------------------------- */
 
 #ifndef LIBCONFIG_STATIC
-#if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__)    \
-     || defined(WIN64) || defined(_WIN64))
+#if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__) \
+  || defined(WIN64) || defined(_WIN64) || defined(__WIN64__))
 
 BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
 {
@@ -108,11 +103,6 @@ static void __config_locale_override(void)
 
 /* ------------------------------------------------------------------------- */
 
-#define __config_has_option(C, O)               \
-  (((C)->options & (O)) != 0)
-
-/* ------------------------------------------------------------------------- */
-
 static void __config_locale_restore(void)
 {
 #if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__)) \
@@ -220,36 +210,10 @@ static void __config_write_value(const config_t *config,
     /* float */
     case CONFIG_TYPE_FLOAT:
     {
-      char *q;
-
-      snprintf(fbuf, sizeof(fbuf) - 3, "%.*f",
-               TAB2DIGITS(config->tab_width),   /* XXX: hack; See config_set_float_precision */
-               value->fval
-      );
-
-      /* check for exponent */
-      q = strchr(fbuf, 'e');
-      if(! q)
-      {
-        /* no exponent */
-        if(! strchr(fbuf, '.')) /* no decimal point */
-          strcat(fbuf, ".0");
-        else
-        {
-          /* has decimal point */
-          char *p;
-
-          for(p = fbuf + strlen(fbuf) - 1; p > fbuf; --p)
-          {
-            if(*p != '0')
-            {
-              *(++p) = '\0';
-              break;
-            }
-          }
-        }
-      }
-
+      const int sci_ok = config_get_option(
+            config, CONFIG_OPTION_ALLOW_SCIENTIFIC_NOTATION);
+      format_double(value->fval, config->float_precision, sci_ok, fbuf,
+                    sizeof(fbuf));
       fputs(fbuf, stream);
       break;
     }
@@ -307,7 +271,7 @@ static void __config_write_value(const config_t *config,
     {
       config_list_t *list = value->list;
 
-      fprintf(stream, "( ");
+      fputs("( ", stream);
 
       if(list)
       {
@@ -336,7 +300,7 @@ static void __config_write_value(const config_t *config,
     {
       config_list_t *list = value->list;
 
-      fprintf(stream, "[ ");
+      fputs("[ ", stream);
 
       if(list)
       {
@@ -367,15 +331,15 @@ static void __config_write_value(const config_t *config,
 
       if(depth > 0)
       {
-        if((config->options & CONFIG_OPTION_OPEN_BRACE_ON_SEPARATE_LINE) != 0)
+        if(config_get_option(config, CONFIG_OPTION_OPEN_BRACE_ON_SEPARATE_LINE))
         {
           fputc('\n', stream);
 
           if(depth > 1)
-            __config_indent(stream, depth, TAB2TAB(config->tab_width));
+            __config_indent(stream, depth, config->tab_width);
         }
 
-        fprintf(stream, "{\n");
+        fputs("{\n", stream);
       }
 
       if(list)
@@ -388,7 +352,7 @@ static void __config_write_value(const config_t *config,
       }
 
       if(depth > 1)
-        __config_indent(stream, depth, TAB2TAB(config->tab_width));
+        __config_indent(stream, depth, config->tab_width);
 
       if(depth > 0)
         fputc('}', stream);
@@ -473,14 +437,12 @@ static void __config_setting_destroy(config_setting_t *setting)
   if(setting)
   {
     if(setting->name)
-      _delete(setting->name);
+      __delete(setting->name);
 
     if(setting->type == CONFIG_TYPE_STRING)
-      _delete(setting->value.sval);
+      __delete(setting->value.sval);
 
-    else if((setting->type == CONFIG_TYPE_GROUP)
-            || (setting->type == CONFIG_TYPE_ARRAY)
-            || (setting->type == CONFIG_TYPE_LIST))
+    else if(config_setting_is_aggregate(setting))
     {
       if(setting->value.list)
         __config_list_destroy(setting->value.list);
@@ -489,7 +451,7 @@ static void __config_setting_destroy(config_setting_t *setting)
     if(setting->hook && setting->config->destructor)
       setting->config->destructor(setting->hook);
 
-    _delete(setting);
+    __delete(setting);
   }
 }
 
@@ -508,37 +470,44 @@ static void __config_list_destroy(config_list_t *list)
     for(p = list->elements, i = 0; i < list->length; p++, i++)
       __config_setting_destroy(*p);
 
-    _delete(list->elements);
+    __delete(list->elements);
   }
 
-  _delete(list);
+  __delete(list);
 }
 
 /* ------------------------------------------------------------------------- */
 
-static int __config_vector_checktype(const config_setting_t *vector, int type)
+static int __config_list_checktype(const config_setting_t *setting, int type)
 {
   /* if the array is empty, then it has no type yet */
 
-  if(! vector->value.list)
+  if(! setting->value.list)
     return(CONFIG_TRUE);
 
-  if(vector->value.list->length == 0)
+  if(setting->value.list->length == 0)
     return(CONFIG_TRUE);
 
   /* if it's a list, any type is allowed */
 
-  if(vector->type == CONFIG_TYPE_LIST)
+  if(setting->type == CONFIG_TYPE_LIST)
     return(CONFIG_TRUE);
 
   /* otherwise the first element added determines the type of the array */
 
-  return((vector->value.list->elements[0]->type == type)
+  return((setting->value.list->elements[0]->type == type)
          ? CONFIG_TRUE : CONFIG_FALSE);
 }
 
 /* ------------------------------------------------------------------------- */
 
+static int __config_type_is_scalar(int type)
+{
+  return((type >= CONFIG_TYPE_INT) && (type <= CONFIG_TYPE_BOOL));
+}
+
+/* ------------------------------------------------------------------------- */
+
 static int __config_validate_name(const char *name)
 {
   const char *p = name;
@@ -568,20 +537,7 @@ static int __config_read(config_t *config, FILE *stream, const char *filename,
   struct parse_context parse_ctx;
   int r;
 
-  /* Reinitialize the config */
-  void (*destructor)(void *) = config->destructor;
-  const char *include_dir = config->include_dir;
-  unsigned short tab_width = config->tab_width;
-  int options = config->options;
-
-  config->include_dir = NULL;
-  config_destroy(config);
-  config_init(config);
-
-  config->destructor = destructor;
-  config->include_dir = include_dir;
-  config->tab_width = tab_width;
-  config->options = options;
+  config_clear(config);
 
   parsectx_init(&parse_ctx);
   parse_ctx.config = config;
@@ -616,7 +572,7 @@ static int __config_read(config_t *config, FILE *stream, const char *filename,
   }
 
   libconfig_yylex_destroy(scanner);
-  config->filenames = scanctx_cleanup(&scan_ctx, &(config->num_filenames));
+  config->filenames = scanctx_cleanup(&scan_ctx);
   parsectx_cleanup(&parse_ctx);
 
   __config_locale_restore();
@@ -644,14 +600,14 @@ static void __config_write_setting(const config_t *config,
                                    const config_setting_t *setting,
                                    FILE *stream, int depth)
 {
-  char group_assign_char = __config_has_option(
+  char group_assign_char = config_get_option(
     config, CONFIG_OPTION_COLON_ASSIGNMENT_FOR_GROUPS) ? ':' : '=';
 
-  char nongroup_assign_char = __config_has_option(
+  char nongroup_assign_char = config_get_option(
     config, CONFIG_OPTION_COLON_ASSIGNMENT_FOR_NON_GROUPS) ? ':' : '=';
 
   if(depth > 1)
-    __config_indent(stream, depth, TAB2TAB(config->tab_width));
+    __config_indent(stream, depth, config->tab_width);
 
 
   if(setting->name)
@@ -667,7 +623,7 @@ static void __config_write_setting(const config_t *config,
 
   if(depth > 0)
   {
-    if(__config_has_option(config, CONFIG_OPTION_SEMICOLON_SEPARATORS))
+    if(config_get_option(config, CONFIG_OPTION_SEMICOLON_SEPARATORS))
       fputc(';', stream);
 
     fputc('\n', stream);
@@ -744,26 +700,29 @@ int config_write_file(config_t *config, const char *filename)
 
 void config_destroy(config_t *config)
 {
-  unsigned int count = config->num_filenames;
-  const char **f;
-
-  __config_setting_destroy(config->root);
-
-  for(f = config->filenames; count > 0; ++f, --count)
-    _delete(*f);
-
-  _delete(config->filenames);
-  _delete(config->include_dir);
-
-  memset((void *)config, 0, sizeof(config_t));
+  config_clear(config);
+  __delete(config->include_dir);
+  __zero(config);
 }
 
 /* ------------------------------------------------------------------------- */
 
-void config_set_tab_width(config_t *config,
-                          unsigned short width)
+void config_clear(config_t *config)
 {
-  /* As per documentation: valid range= 0..15 */
+  /* Destroy the root setting (recursively) and then create a new one. */
+  __config_setting_destroy(config->root);
+  __delete_vec(config->filenames);
+
+  config->root = __new(config_setting_t);
+  config->root->type = CONFIG_TYPE_GROUP;
+  config->root->config = config;
+}
+
+/* ------------------------------------------------------------------------- */
+
+void config_set_tab_width(config_t *config, unsigned short width)
+{
+  /* As per documentation: valid range is 0 - 15. */
   config->tab_width = (width <= 15) ? width : 15;
 }
 
@@ -771,58 +730,37 @@ void config_set_tab_width(config_t *config,
 
 unsigned short config_get_tab_width(const config_t *config)
 {
-  return TAB2TAB(config->tab_width);
+  return config->tab_width;
 }
 
 /* ------------------------------------------------------------------------- */
 
-void config_set_float_precision(config_t *config,
-                                          unsigned short digits)
+void config_set_float_precision(config_t *config, unsigned short digits)
 {
-  config->tab_width |= DIGITS2TAB(digits);
+  config->float_precision = digits;
 }
 
+/* ------------------------------------------------------------------------- */
+
 unsigned short config_get_float_precision(const config_t *config)
 {
-  return TAB2DIGITS(config->tab_width);
+  return config->float_precision;
 }
 
 /* ------------------------------------------------------------------------- */
 
 void config_init(config_t *config)
 {
-  memset((void *)config, 0, sizeof(config_t));
+  __zero(config);
+  config_clear(config);
 
-  config->root = _new(config_setting_t);
-  config->root->type = CONFIG_TYPE_GROUP;
-  config->root->config = config;
+  /* Set default options. */
   config->options = (CONFIG_OPTION_SEMICOLON_SEPARATORS
                      | CONFIG_OPTION_COLON_ASSIGNMENT_FOR_GROUPS
                      | CONFIG_OPTION_OPEN_BRACE_ON_SEPARATE_LINE);
-
-  /* @version 1.6: piggyback float_digits on top of tab_width
-   * (ab)using the existing macros' 0x0F mask in order to preserve
-   * API & ABI compatibility ; changes are fully backwards compatible
-   */
-  config->tab_width = FLOAT_PRECISION;
-  config->tab_width |= DIGITS2TAB(2);    /* float_digits */
-}
-
-/* ------------------------------------------------------------------------- */
-
-void config_set_auto_convert(config_t *config, int flag)
-{
-  if(flag)
-    config->options |= CONFIG_OPTION_AUTOCONVERT;
-  else
-    config->options &= ~CONFIG_OPTION_AUTOCONVERT;
-}
-
-/* ------------------------------------------------------------------------- */
-
-int config_get_auto_convert(const config_t *config)
-{
-  return(__config_has_option(config, CONFIG_OPTION_AUTOCONVERT));
+  config->tab_width = DEFAULT_TAB_WIDTH;
+  config->float_precision = DEFAULT_FLOAT_PRECISION;
+  config->include_fn = config_default_include_func;
 }
 
 /* ------------------------------------------------------------------------- */
@@ -841,18 +779,40 @@ int config_get_options(const config_t *config)
 
 /* ------------------------------------------------------------------------- */
 
+void config_set_option(config_t *config, int option, int flag)
+{
+  if(flag)
+    config->options |= option;
+  else
+    config->options &= ~option;
+}
+
+/* ------------------------------------------------------------------------- */
+
+int config_get_option(const config_t *config, int option)
+{
+  return((config->options & option) == option);
+}
+
+/* ------------------------------------------------------------------------- */
+
+void config_set_hook(config_t *config, void *hook)
+{
+  config->hook = hook;
+}
+
+/* ------------------------------------------------------------------------- */
+
 static config_setting_t *config_setting_create(config_setting_t *parent,
                                                const char *name, int type)
 {
   config_setting_t *setting;
   config_list_t *list;
 
-  if((parent->type != CONFIG_TYPE_GROUP)
-     && (parent->type != CONFIG_TYPE_ARRAY)
-     && (parent->type != CONFIG_TYPE_LIST))
+  if(!config_setting_is_aggregate(parent))
     return(NULL);
 
-  setting = _new(config_setting_t);
+  setting = __new(config_setting_t);
   setting->parent = parent;
   setting->name = (name == NULL) ? NULL : strdup(name);
   setting->type = type;
@@ -863,7 +823,7 @@ static config_setting_t *config_setting_create(config_setting_t *parent,
   list = parent->value.list;
 
   if(! list)
-    list = parent->value.list = _new(config_list_t);
+    list = parent->value.list = __new(config_list_t);
 
   __config_list_add(list, setting);
 
@@ -882,21 +842,23 @@ static int __config_setting_get_int(const config_setting_t *setting,
       return(CONFIG_TRUE);
 
     case CONFIG_TYPE_INT64:
-      if((setting->value.llval > INT32_MAX)
-         || (setting->value.llval < INT32_MIN))
-        *value = 0;
-      else
+      if((setting->value.llval >= INT_MIN)
+         && (setting->value.llval <= INT_MAX))
+      {
         *value = (int)(setting->value.llval);
-      return(CONFIG_TRUE);
+        return(CONFIG_TRUE);
+      }
+      else
+        return(CONFIG_FALSE);
 
     case CONFIG_TYPE_FLOAT:
-      if(__config_has_option(setting->config, CONFIG_OPTION_AUTOCONVERT))
+      if(config_get_option(setting->config, CONFIG_OPTION_AUTOCONVERT))
       {
         *value = (int)(setting->value.fval);
         return(CONFIG_TRUE);
       }
       else
-      { /* fall through */ }
+        return(CONFIG_FALSE);
 
     default:
       return(CONFIG_FALSE);
@@ -928,13 +890,13 @@ static int __config_setting_get_int64(const config_setting_t *setting,
       return(CONFIG_TRUE);
 
     case CONFIG_TYPE_FLOAT:
-      if(__config_has_option(setting->config, CONFIG_OPTION_AUTOCONVERT))
+      if(config_get_option(setting->config, CONFIG_OPTION_AUTOCONVERT))
       {
         *value = (long long)(setting->value.fval);
         return(CONFIG_TRUE);
       }
       else
-      { /* fall through */ }
+        return(CONFIG_FALSE);
 
     default:
       return(CONFIG_FALSE);
@@ -1001,7 +963,7 @@ static int __config_setting_get_float(const config_setting_t *setting,
         return(CONFIG_TRUE);
       }
       else
-      { /* fall through */ }
+        return(CONFIG_FALSE);
 
     default:
       return(CONFIG_FALSE);
@@ -1104,11 +1066,13 @@ int config_setting_set_int64(config_setting_t *setting, long long value)
       return(CONFIG_TRUE);
 
     case CONFIG_TYPE_INT:
-      if((value > INT32_MAX) || (value < INT32_MIN))
-        setting->value.ival = 0;
-      else
+      if((value >= INT_MIN) && (value <= INT_MAX))
+      {
         setting->value.ival = (int)value;
-      return(CONFIG_TRUE);
+        return(CONFIG_TRUE);
+      }
+      else
+        return(CONFIG_FALSE);
 
     case CONFIG_TYPE_FLOAT:
       if(config_get_auto_convert(setting->config))
@@ -1139,7 +1103,7 @@ int config_setting_set_float(config_setting_t *setting, double value)
       return(CONFIG_TRUE);
 
     case CONFIG_TYPE_INT:
-      if(__config_has_option(setting->config, CONFIG_OPTION_AUTOCONVERT))
+      if(config_get_option(setting->config, CONFIG_OPTION_AUTOCONVERT))
       {
         setting->value.ival = (int)value;
         return(CONFIG_TRUE);
@@ -1148,7 +1112,7 @@ int config_setting_set_float(config_setting_t *setting, double value)
         return(CONFIG_FALSE);
 
     case CONFIG_TYPE_INT64:
-      if(__config_has_option(setting->config, CONFIG_OPTION_AUTOCONVERT))
+      if(config_get_option(setting->config, CONFIG_OPTION_AUTOCONVERT))
       {
         setting->value.llval = (long long)value;
         return(CONFIG_TRUE);
@@ -1198,7 +1162,7 @@ int config_setting_set_string(config_setting_t *setting, const char *value)
     return(CONFIG_FALSE);
 
   if(setting->value.sval)
-    _delete(setting->value.sval);
+    __delete(setting->value.sval);
 
   setting->value.sval = (value == NULL) ? NULL : strdup(value);
   return(CONFIG_TRUE);
@@ -1336,33 +1300,34 @@ int config_lookup_bool(const config_t *config, const char *path, int *value)
 
 /* ------------------------------------------------------------------------- */
 
-int config_setting_get_int_elem(const config_setting_t *vector, int idx)
+int config_setting_get_int_elem(const config_setting_t *setting, int idx)
 {
-  const config_setting_t *element = config_setting_get_elem(vector, idx);
+  const config_setting_t *element = config_setting_get_elem(setting, idx);
 
   return(element ? config_setting_get_int(element) : 0);
 }
 
 /* ------------------------------------------------------------------------- */
 
-config_setting_t *config_setting_set_int_elem(config_setting_t *vector,
+config_setting_t *config_setting_set_int_elem(config_setting_t *setting,
                                               int idx, int value)
 {
   config_setting_t *element = NULL;
 
-  if((vector->type != CONFIG_TYPE_ARRAY) && (vector->type != CONFIG_TYPE_LIST))
+  if((setting->type != CONFIG_TYPE_ARRAY)
+     && (setting->type != CONFIG_TYPE_LIST))
     return(NULL);
 
   if(idx < 0)
   {
-    if(! __config_vector_checktype(vector, CONFIG_TYPE_INT))
+    if(! __config_list_checktype(setting, CONFIG_TYPE_INT))
       return(NULL);
 
-    element = config_setting_create(vector, NULL, CONFIG_TYPE_INT);
+    element = config_setting_create(setting, NULL, CONFIG_TYPE_INT);
   }
   else
   {
-    element = config_setting_get_elem(vector, idx);
+    element = config_setting_get_elem(setting, idx);
 
     if(! element)
       return(NULL);
@@ -1376,34 +1341,35 @@ config_setting_t *config_setting_set_int_elem(config_setting_t *vector,
 
 /* ------------------------------------------------------------------------- */
 
-long long config_setting_get_int64_elem(const config_setting_t *vector,
+long long config_setting_get_int64_elem(const config_setting_t *setting,
                                         int idx)
 {
-  const config_setting_t *element = config_setting_get_elem(vector, idx);
+  const config_setting_t *element = config_setting_get_elem(setting, idx);
 
   return(element ? config_setting_get_int64(element) : 0);
 }
 
 /* ------------------------------------------------------------------------- */
 
-config_setting_t *config_setting_set_int64_elem(config_setting_t *vector,
+config_setting_t *config_setting_set_int64_elem(config_setting_t *setting,
                                                 int idx, long long value)
 {
   config_setting_t *element = NULL;
 
-  if((vector->type != CONFIG_TYPE_ARRAY) && (vector->type != CONFIG_TYPE_LIST))
+  if((setting->type != CONFIG_TYPE_ARRAY)
+     && (setting->type != CONFIG_TYPE_LIST))
     return(NULL);
 
   if(idx < 0)
   {
-    if(! __config_vector_checktype(vector, CONFIG_TYPE_INT64))
+    if(! __config_list_checktype(setting, CONFIG_TYPE_INT64))
       return(NULL);
 
-    element = config_setting_create(vector, NULL, CONFIG_TYPE_INT64);
+    element = config_setting_create(setting, NULL, CONFIG_TYPE_INT64);
   }
   else
   {
-    element = config_setting_get_elem(vector, idx);
+    element = config_setting_get_elem(setting, idx);
 
     if(! element)
       return(NULL);
@@ -1417,32 +1383,33 @@ config_setting_t *config_setting_set_int64_elem(config_setting_t *vector,
 
 /* ------------------------------------------------------------------------- */
 
-double config_setting_get_float_elem(const config_setting_t *vector, int idx)
+double config_setting_get_float_elem(const config_setting_t *setting, int idx)
 {
-  config_setting_t *element = config_setting_get_elem(vector, idx);
+  config_setting_t *element = config_setting_get_elem(setting, idx);
 
   return(element ? config_setting_get_float(element) : 0.0);
 }
 
 /* ------------------------------------------------------------------------- */
 
-config_setting_t *config_setting_set_float_elem(config_setting_t *vector,
+config_setting_t *config_setting_set_float_elem(config_setting_t *setting,
                                                 int idx, double value)
 {
   config_setting_t *element = NULL;
 
-  if((vector->type != CONFIG_TYPE_ARRAY) && (vector->type != CONFIG_TYPE_LIST))
+  if((setting->type != CONFIG_TYPE_ARRAY)
+     && (setting->type != CONFIG_TYPE_LIST))
     return(NULL);
 
   if(idx < 0)
   {
-    if(! __config_vector_checktype(vector, CONFIG_TYPE_FLOAT))
+    if(! __config_list_checktype(setting, CONFIG_TYPE_FLOAT))
       return(NULL);
 
-    element = config_setting_create(vector, NULL, CONFIG_TYPE_FLOAT);
+    element = config_setting_create(setting, NULL, CONFIG_TYPE_FLOAT);
   }
   else
-    element = config_setting_get_elem(vector, idx);
+    element = config_setting_get_elem(setting, idx);
 
   if(! element)
     return(NULL);
@@ -1455,9 +1422,9 @@ config_setting_t *config_setting_set_float_elem(config_setting_t *vector,
 
 /* ------------------------------------------------------------------------- */
 
-int config_setting_get_bool_elem(const config_setting_t *vector, int idx)
+int config_setting_get_bool_elem(const config_setting_t *setting, int idx)
 {
-  config_setting_t *element = config_setting_get_elem(vector, idx);
+  config_setting_t *element = config_setting_get_elem(setting, idx);
 
   if(! element)
     return(CONFIG_FALSE);
@@ -1470,23 +1437,24 @@ int config_setting_get_bool_elem(const config_setting_t *vector, int idx)
 
 /* ------------------------------------------------------------------------- */
 
-config_setting_t *config_setting_set_bool_elem(config_setting_t *vector,
+config_setting_t *config_setting_set_bool_elem(config_setting_t *setting,
                                                int idx, int value)
 {
   config_setting_t *element = NULL;
 
-  if((vector->type != CONFIG_TYPE_ARRAY) && (vector->type != CONFIG_TYPE_LIST))
+  if((setting->type != CONFIG_TYPE_ARRAY)
+     && (setting->type != CONFIG_TYPE_LIST))
     return(NULL);
 
   if(idx < 0)
   {
-    if(! __config_vector_checktype(vector, CONFIG_TYPE_BOOL))
+    if(! __config_list_checktype(setting, CONFIG_TYPE_BOOL))
       return(NULL);
 
-    element = config_setting_create(vector, NULL, CONFIG_TYPE_BOOL);
+    element = config_setting_create(setting, NULL, CONFIG_TYPE_BOOL);
   }
   else
-    element = config_setting_get_elem(vector, idx);
+    element = config_setting_get_elem(setting, idx);
 
   if(! element)
     return(NULL);
@@ -1499,10 +1467,10 @@ config_setting_t *config_setting_set_bool_elem(config_setting_t *vector,
 
 /* ------------------------------------------------------------------------- */
 
-const char *config_setting_get_string_elem(const config_setting_t *vector,
+const char *config_setting_get_string_elem(const config_setting_t *setting,
                                            int idx)
 {
-  config_setting_t *element = config_setting_get_elem(vector, idx);
+  config_setting_t *element = config_setting_get_elem(setting, idx);
 
   if(! element)
     return(NULL);
@@ -1515,23 +1483,24 @@ const char *config_setting_get_string_elem(const config_setting_t *vector,
 
 /* ------------------------------------------------------------------------- */
 
-config_setting_t *config_setting_set_string_elem(config_setting_t *vector,
+config_setting_t *config_setting_set_string_elem(config_setting_t *setting,
                                                  int idx, const char *value)
 {
   config_setting_t *element = NULL;
 
-  if((vector->type != CONFIG_TYPE_ARRAY) && (vector->type != CONFIG_TYPE_LIST))
+  if((setting->type != CONFIG_TYPE_ARRAY)
+     && (setting->type != CONFIG_TYPE_LIST))
     return(NULL);
 
   if(idx < 0)
   {
-    if(! __config_vector_checktype(vector, CONFIG_TYPE_STRING))
+    if(! __config_list_checktype(setting, CONFIG_TYPE_STRING))
       return(NULL);
 
-    element = config_setting_create(vector, NULL, CONFIG_TYPE_STRING);
+    element = config_setting_create(setting, NULL, CONFIG_TYPE_STRING);
   }
   else
-    element = config_setting_get_elem(vector, idx);
+    element = config_setting_get_elem(setting, idx);
 
   if(! element)
     return(NULL);
@@ -1544,14 +1513,16 @@ config_setting_t *config_setting_set_string_elem(config_setting_t *vector,
 
 /* ------------------------------------------------------------------------- */
 
-config_setting_t *config_setting_get_elem(const config_setting_t *vector,
+config_setting_t *config_setting_get_elem(const config_setting_t *setting,
                                           unsigned int idx)
-{
-  config_list_t *list = vector->value.list;
+{  
+  config_list_t *list;
 
-  if(((vector->type != CONFIG_TYPE_ARRAY)
-      && (vector->type != CONFIG_TYPE_LIST)
-      && (vector->type != CONFIG_TYPE_GROUP)) || ! list)
+  if(! config_setting_is_aggregate(setting))
+    return(NULL);
+
+  list = setting->value.list;
+  if(! list)
     return(NULL);
 
   if(idx >= list->length)
@@ -1582,17 +1553,22 @@ void config_set_destructor(config_t *config, void (*destructor)(void *))
 
 void config_set_include_dir(config_t *config, const char *include_dir)
 {
-  _delete(config->include_dir);
+  __delete(config->include_dir);
   config->include_dir = strdup(include_dir);
 }
 
 /* ------------------------------------------------------------------------- */
 
+void config_set_include_func(config_t *config, config_include_fn_t func)
+{
+  config->include_fn = func ? func : config_default_include_func;
+}
+
+/* ------------------------------------------------------------------------- */
+
 int config_setting_length(const config_setting_t *setting)
 {
-  if((setting->type != CONFIG_TYPE_GROUP)
-     && (setting->type != CONFIG_TYPE_ARRAY)
-     && (setting->type != CONFIG_TYPE_LIST))
+  if(! config_setting_is_aggregate(setting))
     return(0);
 
   if(! setting->value.list)
@@ -1619,6 +1595,9 @@ config_setting_t *config_setting_add(config_setting_t *parent,
   if(! parent)
     return(NULL);
 
+  if((parent->type == CONFIG_TYPE_ARRAY) && !__config_type_is_scalar(type))
+    return(NULL); /* only scalars can be added to arrays */
+
   if((parent->type == CONFIG_TYPE_ARRAY) || (parent->type == CONFIG_TYPE_LIST))
     name = NULL;
 
@@ -1650,7 +1629,7 @@ int config_setting_remove(config_setting_t *parent, const char *name)
     return(CONFIG_FALSE);
 
   setting = config_setting_lookup(parent, name);
-  if(!setting)
+  if(! setting)
     return(CONFIG_FALSE);
 
   settingName = name;
@@ -1687,11 +1666,11 @@ int config_setting_remove_elem(config_setting_t *parent, unsigned int idx)
   if(! parent)
     return(CONFIG_FALSE);
 
-  list = parent->value.list;
+  if(! config_setting_is_aggregate(parent))
+    return(CONFIG_FALSE);
 
-  if(((parent->type != CONFIG_TYPE_ARRAY)
-      && (parent->type != CONFIG_TYPE_LIST)
-      && (parent->type != CONFIG_TYPE_GROUP)) || ! list)
+  list = parent->value.list;
+  if(! list)
     return(CONFIG_FALSE);
 
   if(idx >= list->length)
@@ -1727,3 +1706,47 @@ int config_setting_index(const config_setting_t *setting)
 
 /* ------------------------------------------------------------------------- */
 
+const char **config_default_include_func(config_t *config,
+                                         const char *include_dir,
+                                         const char *path,
+                                         const char **error)
+{
+  char *file;
+  const char **files;
+
+  if(include_dir && IS_RELATIVE_PATH(path))
+  {
+    file = (char *)malloc(strlen(include_dir) + strlen(path) + 2);
+    strcpy(file, include_dir);
+    strcat(file, FILE_SEPARATOR);
+    strcat(file, path);
+  }
+  else
+    file = strdup(path);
+
+  *error = NULL;
+
+  files = (const char **)malloc(sizeof(char **) * 2);
+  files[0] = file;
+  files[1] = NULL;
+
+  return(files);
+}
+
+/* ------------------------------------------------------------------------- */
+
+int config_setting_is_scalar(const config_setting_t *setting)
+{
+  return(__config_type_is_scalar(setting->type));
+}
+
+/* ------------------------------------------------------------------------- */
+
+int config_setting_is_aggregate(const config_setting_t *setting)
+{
+  return((setting->type == CONFIG_TYPE_ARRAY)
+         || (setting->type == CONFIG_TYPE_LIST)
+         || (setting->type == CONFIG_TYPE_GROUP));
+}
+
+/* ------------------------------------------------------------------------- */
