@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------------
    libconfig - A library for processing structured configuration files
-   Copyright (C) 2005-2015  Mark A Lindner
+   Copyright (C) 2005-2018  Mark A Lindner
 
    This file is part of libconfig.
 
@@ -23,7 +23,17 @@
 #ifndef __wincompat_h
 #define __wincompat_h
 
-#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__)
+#include <limits.h>
+
+#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) \
+  || defined(WIN64) || defined(_WIN64) || defined(__WIN64__)
+
+/* Prevent warnings about redefined malloc/free in generated code. */
+#ifndef _STDLIB_H
+#define _STDLIB_H
+#endif
+
+#include <malloc.h>
 
 #ifdef _MSC_VER
 #pragma warning (disable: 4996)
@@ -33,9 +43,10 @@
 #include <windows.h>
 
 #define fileno _fileno
-#define fstat _fstat
-#define stat _stat // struct stat for fstat()
+
+#if _MSC_VER < 1800
 #define snprintf  _snprintf
+#endif
 
 #if !defined(__MINGW32__) && _MSC_VER < 1800
 #define atoll     _atoi64
@@ -50,11 +61,8 @@
 #endif
 
 #if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__) \
-     || defined(__MINGW32__))
-
-/* Why does gcc on MinGW use the Visual C++ style format directives
- * for 64-bit integers? Inquiring minds want to know....
- */
+  || defined(WIN64) || defined(_WIN64) || defined(__WIN64__) \
+  || defined(__MINGW32__))
 
 #define INT64_FMT "%I64d"
 #define UINT64_FMT "%I64u"
@@ -74,25 +82,33 @@
 
 #endif /* defined(WIN32) || defined(__MINGW32__) */
 
-#if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__)) \
+#if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__) \
+  || defined(WIN64) || defined(_WIN64) || defined(__WIN64__)) \
   && ! defined(__MINGW32__)
 
 #define INT64_CONST(I)  (I ## i64)
 #define UINT64_CONST(I) (I ## Ui64)
 
-#ifndef INT32_MAX
-#define INT32_MAX (2147483647)
+#ifndef INT_MAX
+#define INT_MAX (2147483647)
 #endif
 
-#ifndef INT32_MIN
-#define INT32_MIN (-2147483647-1)
+#ifndef INT_MIN
+#define INT_MIN (-2147483647-1)
 #endif
 
-#else /* defined(WIN32) && ! defined(__MINGW32__) */
+#include <Shlwapi.h>
+#define IS_RELATIVE_PATH(P) \
+  (PathIsRelativeA(P))
+
+#else /* defined(WIN32/WIN64) && ! defined(__MINGW32__) */
 
 #define INT64_CONST(I)  (I ## LL)
 #define UINT64_CONST(I) (I ## ULL)
 
-#endif /* defined(WIN32) && ! defined(__MINGW32__) */
+#define IS_RELATIVE_PATH(P) \
+  ((P)[0] != '/')
+
+#endif /* defined(WIN32/WIN64) && ! defined(__MINGW32__) */
 
 #endif /* __wincompat_h */
